# -*- coding: future_fstrings -*-

import curses
import curses.textpad
from math import floor

DEFAULT_MOTOR_KEYS = ('q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '[', ']')
DEFAULT_SERVO_KEYS = ('a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', "'", '#')
DEFAULT_ADC_KEYS = ('z', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', '/')
DEFAULT_TITLE = 'Approxeng.hwsupport console by @Approx_Eng'

VALUE_KEYS = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-']
VALUE_ITEMS = list([round(i / ((len(VALUE_KEYS) - 1) / 2) - 1, 1) for i in range(0, len(VALUE_KEYS))])


def run_curses_gui(board, motor_keys=DEFAULT_MOTOR_KEYS, servo_keys=DEFAULT_SERVO_KEYS, adc_keys=DEFAULT_ADC_KEYS,
                   title=DEFAULT_TITLE):
    """
    Build and run the GUI for the supplied board. See build_curses_gui for parameter documentation
    """
    curses.wrapper(build_curses_gui(board, motor_keys, servo_keys, adc_keys, title))


def build_curses_gui(board, motor_keys=DEFAULT_MOTOR_KEYS, servo_keys=DEFAULT_SERVO_KEYS, adc_keys=DEFAULT_ADC_KEYS,
                     title=DEFAULT_TITLE):
    """
    Build the GUI for a given board instance.
    :param board:
        A board which must have been augmented by the approxeng.hwsupport module
    :param motor_keys:
        Keys used to select motor driver channels, defaults to the QWERTYUIOP[] row on the keyboard. Must contain at
        least as many items as you have motor driver channels, only those corresponding to motor drivers on the supplied
        board will be shown.
    :param servo_keys:
        Keys used to select servos, defaults to the ASDFGHKL;'# row on the keyboard. Must contain at least as many items
        as you have servos, only those corresponding to servos on the supplied board will be shown.
    :param adc_keys:
        Keys used to select ADC channels, defaults to ZXCVBNM,./ row on the keyboard. Must contain at least as many
        items as you have ADCs, only those corresponding to ADCs on the supplied board will be shown.
    :param title:
        Shown on the first line of the generated console
    :return:
        A function which can be used as the argument to curses.wrapper(..) to show the GUI
    """

    def curses_main(screen):
        try:
            display = DisplayState(screen=screen, board=board, motor_keys=motor_keys, servo_keys=servo_keys,
                                   adc_keys=adc_keys)
            curses.cbreak()
            curses.halfdelay(1)
            while True:
                display.start()

                # Draw title
                display.println(title)
                display.println('Letters to select control, numbers to set value, SPACE stops all, CTRL-C to exit')
                display.newline()

                # Motors, if present
                if display.motors:
                    display.print_header('Motors')
                    for index, motor in enumerate(display.motors):
                        row, col = divmod(index, 4)
                        display.show_motor(display.line + row, col * 20, motor)
                    display.line += floor((len(display.motors) - 1) / 4) + 1
                    display.newline()

                # Servos, if present
                if display.servo_pins:
                    display.print_header('Servos')
                    for index, servo in enumerate(display.servo_pins):
                        row, col = divmod(index, 4)
                        display.show_servo(display.line + row, col * 20, servo)
                    display.line += floor((len(display.servo_pins) - 1) / 4) + 1
                    display.newline()

                # ADC channels, if present
                if display.adcs:
                    display.print_header('ADC Channels')
                    for index, adc in enumerate(display.adcs):
                        row, col = divmod(index, 4)
                        display.show_adc(display.line + row, col * 20, adc)
                    display.line += floor((len(display.adcs) - 1) / 4) + 1
                    display.newline()

                # Show values if either motor or servo, doesn't make any sense for e.g. ADCs
                if display.control_is_servo or display.control_is_motor:
                    display.println('Values - number key row or up / down arrows to set, BACKSPACE to stop / disable')
                    for value in VALUE_ITEMS:
                        index = VALUE_ITEMS.index(value)
                        display.show_value(display.line + len(VALUE_ITEMS) - index - 1, 2, value)

                # Show editor if the selected control has one
                if display.control_is_servo:
                    editor = ServoConfigEditor(display=display, row=display.line + 1, column=40, height=4)
                    editor.render()
                elif display.control_is_motor:
                    editor = MotorConfigEditor(display=display, row=display.line + 1, column=40, height=3)
                    editor.render()
                elif display.control_is_adc:
                    editor = ADCConfigEditor(display=display, row=display.line, column=40, height=3)
                    editor.render()
                else:
                    editor = None

                # Wait for a keypress and respond to it
                try:
                    key = screen.getkey()
                    if key in motor_keys and motor_keys.index(key) < board.num_motors:
                        display.control = f'm{display.motors[motor_keys.index(key)]}'
                    elif key in servo_keys and servo_keys.index(key) < len(display.servo_pins):
                        display.control = f's{display.servo_pins[servo_keys.index(key)]}'
                    elif key in adc_keys and adc_keys.index(key) < len(display.adcs):
                        display.control = f'adc{display.adcs[adc_keys.index(key)]}'
                    elif key == ' ':
                        board.stop()
                    elif key == 'KEY_LEFT':
                        display.select_previous_control()
                    elif key == 'KEY_RIGHT':
                        display.select_next_control()
                    elif key == '=' and editor is not None:
                        editor.edit()
                        curses.cbreak()
                        curses.halfdelay(1)
                    elif key == 'KEY_BACKSPACE':
                        if display.control_is_servo:
                            display.value = None
                        elif display.control_is_motor:
                            display.value = 0
                    # Setting value is only meaningful if we've got a servo or motor selected
                    if display.control_is_servo or display.control_is_motor:
                        if key in VALUE_KEYS:
                            display.value = VALUE_ITEMS[VALUE_KEYS.index(key)]
                        elif key == 'KEY_UP':
                            display.value = round(min(display.value + 0.2, 1.0), 1) if display.value is not None else 0
                        elif key == 'KEY_DOWN':
                            display.value = round(max(display.value - 0.2, -1.0), 1) if display.value is not None else 0
                except curses.error:
                    # No input available, perfectly normal
                    pass
        except KeyboardInterrupt:
            # Exit on CTRL-C, stopping the motors as we go
            board.stop()

    return curses_main


class DisplayState:

    def __init__(self, screen, board, motor_keys, servo_keys, adc_keys):
        self.screen = screen

        self.line = 0
        self.servo_pins = board.servos
        self.board = board
        self.adcs = board.adcs
        self.motors = board.motors
        self.motor_keys = motor_keys
        self.servo_keys = servo_keys
        self.adc_keys = adc_keys
        self.all_controls = list([f'm{motor}' for motor in self.motors]) + list(
            [f's{servo}' for servo in self.servo_pins]) + list([f'adc{adc}' for adc in self.adcs])
        self.control = self.all_controls[0]
        # Disable echo to terminal
        curses.noecho()
        # Hide the cursor
        curses.curs_set(0)
        # Contrast colour for UI
        curses.init_pair(1, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        # Highlight
        curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_WHITE)
        # Enable colour
        curses.start_color()
        # Clear the screen
        screen.clear()
        # Enable key events for special keys i.e. arrows, backspace
        screen.keypad(True)

    def start(self):
        self.screen.clear()
        self.line = 0

    @property
    def value(self):
        return self.board.__getattribute__(self.control)

    @value.setter
    def value(self, value):
        self.board.__setattr__(self.control, value)

    def println(self, string, contrast=False):
        try:
            if contrast:
                self.screen.addstr(self.line, 0, string, curses.color_pair(1))
            else:
                self.screen.addstr(self.line, 0, string)
        except curses.error:
            pass
        self.line += 1

    def select_next_control(self):
        control_index = self.all_controls.index(self.control)
        self.control = self.all_controls[(control_index + 1) % len(self.all_controls)]

    def select_previous_control(self):
        control_index = self.all_controls.index(self.control)
        self.control = self.all_controls[(control_index - 1) % len(self.all_controls)]

    def newline(self):
        self.line += 1

    def print_header(self, string):
        s = '——' + string
        s += '—' * (80 - len(s))
        self.println(s, True)

    def show_motor(self, row, col, motor):
        try:
            try:
                speed = self.board.__getattribute__(f'm{motor}')
            except AttributeError:
                speed = None
            speed_string = '??' if speed is None else f'{speed:.1f}'
            rep = f'm{motor}[{self.motor_keys[motor]}] = {speed_string}'
            if self.control == f'm{motor}':
                self.screen.addstr(row, col, rep, curses.color_pair(2))
            else:
                self.screen.addstr(row, col, rep)
        except curses.error:
            pass

    @property
    def control_is_servo(self):
        return self.control[:1] == 's'

    @property
    def control_is_motor(self):
        return self.control[:1] == 'm'

    @property
    def control_is_adc(self):
        return self.control[:1] == 'a'

    def show_servo(self, row, col, servo):
        try:
            try:
                value = self.board.__getattribute__(f's{servo}')
            except AttributeError:
                value = None
            value_string = '--' if value is None else f'{value:.1f}'
            rep = f's{servo:02}[{self.servo_keys[self.servo_pins.index(servo)]}] = {value_string}'
            if self.control == f's{servo}':
                self.screen.addstr(row, col, rep, curses.color_pair(2))
            else:
                self.screen.addstr(row, col, rep)
        except curses.error:
            pass

    def show_adc(self, row, col, adc):
        try:
            try:
                value = self.board.__getattribute__(f'adc{adc}')
            except AttributeError:
                value = None
            value_string = '--' if value is None else f'{value:.2f}'
            rep = f'adc{adc:01}[{self.adc_keys[self.adcs.index(adc)]}] = {value_string}v'
            if self.control == f'adc{adc}':
                self.screen.addstr(row, col, rep, curses.color_pair(2))
            else:
                self.screen.addstr(row, col, rep)
        except curses.error:
            pass

    def show_value(self, row, col, value):
        try:
            current_value = self.value
            string = f'[{VALUE_KEYS[VALUE_ITEMS.index(round(value, 1))]}]={round(value, 1)}'
            if current_value is not None and round(current_value, 1) == round(value, 1):
                self.screen.addstr(row, col, string, curses.color_pair(2))
            else:
                self.screen.addstr(row, col, string)
        except curses.error:
            pass


class MotorConfigEditor:
    def __init__(self, display, row, column, height):
        self.display = display
        self.row = row
        self.column = column
        self.height = height

    def render(self):
        try:
            screen = self.display.screen
            curses.textpad.rectangle(screen, self.row, self.column, self.row + self.height, 79)
            screen.addstr(self.row + 1, self.column + 1, f'Motor {self.display.control}, \'=\' to toggle invert:',
                          curses.color_pair(1))
            invert = self.display.board.__getattribute__(f'{self.display.control}_invert')
            screen.addstr(self.row + 2, self.column + 1, f'Invert direction = {invert}')
        except curses.error:
            pass

    def edit(self):
        invert = self.display.board.__getattribute__(f'{self.display.control}_invert')
        self.display.board.__setattr__(f'{self.display.control}_invert', not invert)


class ADCConfigEditor:
    def __init__(self, display, row, column, height):
        self.display = display
        self.row = row
        self.column = column
        self.height = height

    def render(self):
        try:
            screen = self.display.screen
            curses.textpad.rectangle(screen, self.row, self.column, self.row + self.height, 79)
            screen.addstr(self.row + 1, self.column + 1, f'ADC {self.display.control}, \'=\' to calibrate:',
                          curses.color_pair(1))
            divisor = self.display.board.__getattribute__(f'{self.display.control}_divisor')
            screen.addstr(self.row + 2, self.column + 1, f'Current divisor = {divisor:.1f}')
        except curses.error:
            pass

    def edit(self):
        try:
            screen = self.display.screen
            screen.addstr(self.row + 1, self.column + 1, f'Enter observed voltage, then RETURN:  ',
                          curses.color_pair(1))
            screen.addstr(self.row + 2, self.column + 1, f'Measured voltage =                    ')
            curses.echo()
            curses.curs_set(2)
            measured_voltage = screen.getstr(self.row + 2, self.column + 20, 10)

            parsed_measured_voltage = None
            try:
                parsed_measured_voltage = float(measured_voltage)
            except ValueError:
                pass
            if parsed_measured_voltage:
                current_voltage = self.display.value
                current_divisor = self.display.board.__getattribute__(f'{self.display.control}_divisor')
                new_divisor = current_divisor * (current_voltage / parsed_measured_voltage)
                self.display.board.__setattr__(f'{self.display.control}_divisor', new_divisor)
            curses.noecho()
            curses.curs_set(0)
        except curses.error:
            curses.noecho()
            curses.curs_set(0)


class ServoConfigEditor:
    def __init__(self, display, row, column, height):
        self.display = display
        self.row = row
        self.column = column
        self.height = height

    def render(self):
        try:
            screen = self.display.screen
            curses.textpad.rectangle(screen, self.row, self.column, self.row + self.height, 79)
            screen.addstr(self.row + 1, self.column + 1, f'Servo {self.display.control}, \'=\' to edit config:',
                          curses.color_pair(1))
            pulse_min, pulse_max = self.display.board.__getattribute__(f'{self.display.control}_config')
            screen.addstr(self.row + 2, self.column + 1, f'Min pulse width = {pulse_min} μs')
            screen.addstr(self.row + 3, self.column + 1, f'Max pulse width = {pulse_max} μs')
        except curses.error:
            pass

    def edit(self):
        try:
            screen = self.display.screen
            screen.addstr(self.row + 1, self.column + 1, f'Enter new Pulse Min, then RETURN:   ',
                          curses.color_pair(1))
            screen.addstr(self.row + 2, self.column + 1, f'Min pulse width =                 ')
            curses.echo()
            curses.curs_set(2)
            new_min = screen.getstr(self.row + 2, self.column + 19, 10)
            screen.addstr(self.row + 1, self.column + 1, f'Enter new Pulse Max, then RETURN:',
                          curses.color_pair(1))
            screen.addstr(self.row + 3, self.column + 1, f'Max pulse width =                 ')
            new_max = screen.getstr(self.row + 3, self.column + 19, 10)
            parsed_new_min = None
            parsed_new_max = None
            try:
                parsed_new_min = int(new_min)
            except ValueError:
                pass
            try:
                parsed_new_max = int(new_max)
            except ValueError:
                pass
            self.display.board.__setattr__(f'{self.display.control}_config', (parsed_new_min, parsed_new_max))
            curses.noecho()
            curses.curs_set(0)
        except curses.error:
            curses.noecho()
            curses.curs_set(0)



# -*- coding: future_fstrings -*-

import logging

import yaml
from approxeng.hwsupport.adcs import ADCS, ADC, ReadADCsMixin
from approxeng.hwsupport.motors import MOTORS, Motor, SetMotorsMixin
from approxeng.hwsupport.servos import SERVOS, Servo, SetServosMixin
from approxeng.hwsupport.leds import LEDS, LED, SetLEDsMixin

LOGGER = logging.getLogger(name='approxeng.hwsupport')


def add_properties(board, motors=None, servos=None, adcs=None, default_adc_divisor=7891, leds=None):
    """
    Augment an existing instance of a motor, servo, adc, or combination driver class. This wraps up any provided
    methods in ones which check their input ranges properly, exposes those as properties (read and write), adds
    a set of configuration load and store methods, and also adds some generic functionality such as a stop() method.

    The exact properties and methods injected depend on the driver object passed in, and on the supplied lists of
    motors, servos, and adc channel numbers:

    For motors, the underlying board must provide a method _set_motor_speed(motor_index, speed) which accepts an integer
    motor number and a floating point value from -1.0 to 1.0 for speed. If this method exists, and there are items in
    the 'motors' parameter, then the following are added to the driver object:

    1. A new method set_motor_speed injected from the SetMotorsMixin. This takes the same arguments as the original
    _set_motor_speed method, but does value checking on both the motor index and the speed, and in addition tracks any
    requested speed changes to allow them to be read via property access.
    2. For each motor, a set of properties mXX and motorXX which allow write and read of that motor's speed value.
    3. For each motor, a set of properties mXX_invert and motorXX_invert which allow for negation of all values sent
    to subsequent motor speed calls - handy if you've not quite got your wiring right first time.

    For servos, the underlying board must provide a method _set_servo_pulsewidth(servo, pulse_width) accepting an int
    servo index and a desired pulse width specified in microseconds. If this method exists, and there are items in the
    'servos' parameter, the following are added to the driver object:

    1. A new method, set_servo(servo, position), taking the integer index of a servo and a position in the range -1.0
    to 1.0. This method has error checking on the servo index, and will clamp its input position to the -1.0 to 1.0
    range silently.
    2. A new method, disable_servo(servo) which sets the pulse width of the given servo to 0.
    3. For each servo, a set of properties sXX and servoXX which allow read and write access to the servo position. If
    a servo is disabled these will read as None.
    4. For each servo, a set of properties sXX_config and servoXX_config which read and write a pair of values. These
    values are the minimum and maximum pulse widths accepted by the servo on this channel. These default to 500,2500
    unless otherwise set, and are used to interpret the position value of the servo when converting to a pulse width
    for the corresponding output.

    For ADC channels, the underlying board must provide a method _read_adc(adc) accepting an integer adc channel number
    and returning a raw ADC value. If this method exists and there are items in the 'adc' parameter, the following are
    added to the driver object:

    1. A new method, read_adc(adc), taking the same integer adc index as the underlying _read_adc method, and returning
    a floating point voltage. This voltage is calculated by dividing the raw read value by a per-adc-channel divisor.
    2. For each channel, an adcXX property which exposes the value for that channel.
    3. For each channel, an adcXX_divisor property which can be written and read to set and get the per-channel divisor

    For LEDs, the underlying board must provide a method _set_led_rgb(led, r, g, b) taking RGB values as floats from 0.0
    to 1.0. If this method exists and there are entries in the 'leds' parameter, the following are added to the driver
    object:

    1. A new method, set_led_hsv(led, h, s, v) which does range checking on its hue, saturation, and value inputs - hue
    is treated as a continuous quantity so you can set it to anything, i.e. 1.0 is red, as is 0.0, as is 3.0 or -6.
    Saturation and value are clamped to 0.0-1.0
    2. A new method, set_led_brightness(led, brightness) which takes a brightness value from 0.0-1.0 and uses it to
    scale the value part of any colours set subsequently. Handy for when you don't want to blind yourself. Brightness
    is initially 1.0
    3. For each LED, a read / write property ledXX, this can be written to with either a tuple (h, s, v) or a colour
    name from the CSS4 standard colours. Reading the property will always return an HSV tuple
    4. For each LED, a read / write property ledXX_rgb, this can be written to with a tuple of 0.0-1.0 rgb values, and
    will return the same. Internally this is converted to HSV and scaled by the brightness, so your actual RGB values
    pushed to the LED may be different to those returned from this call if brightness is not 1.0
    5. For each LED, a read / write property ledXX_brightness which can be used to set the brightness for that LED from
    0.0 to 1.0


    Configuration properties are also injected, specifically a read / write property 'config' which contains the entire
    configuration for all motors, servos, and adc channels as a dict - writing to this property will only set values
    which exist both in the target object and in the supplied dict, it will ignore any properties not available to the
    object and does not require all properties to be set in one go. In addition a property 'config_yaml' allows the
    current configuration to be read into a YAML string.

    A method 'stop()' is also injected, this will set any motor speeds to zero, disable any servos, and then, if
    provided by the original object, call a '_stop()' function.

    Note - all injected methods take an optional **kwargs argument which will be passed through to the underlying
    object's methods.

    :param board:
        An instance of a motor, servo, adc or combo board object to be augmented. This object will be modified in place
    :param motors:
        An array of integer motor numbers to be exposed for this board, defaults to None for no motors
    :param servos:
        An array of integer servo numbers to be exposed for this board, defaults to None for no servos
    :param adcs:
        An array of integer ADC channel numbers to be exposed for this board, defaults to None for no ADC channels
    :param default_adc_divisor:
        Initial value for all ADC divisor configs, defaults to 7891
    """

    # Replace default values with empty lists
    if motors is None:
        motors = []
    if servos is None:
        servos = []
    if adcs is None:
        adcs = []
    if leds is None:
        leds = []

    # Construct a set of superclasses, applying mixins for each of motors, servos, and adc channels where present
    superclasses = [board.__class__]
    if callable(getattr(board, '_set_motor_speed', None)) and motors:
        superclasses += [SetMotorsMixin]
    if callable(getattr(board, '_set_servo_pulsewidth', None)) and servos:
        superclasses += [SetServosMixin]
    if callable(getattr(board, '_read_adc', None)) and adcs:
        superclasses += [ReadADCsMixin]
    if callable(getattr(board, '_set_led_rgb', None)) and leds:
        superclasses += [SetLEDsMixin]

    class Board(*superclasses):
        """
        There's exactly one instance of this class, it's created dynamically when augmenting an object with the
        motor, servo, and ADC properties.
        """

        def stop(self, **kwargs):
            """
            Used to stop all activity on a board.

            If there are servos, these are disabled. If there are motors, they are set to 0 speed. Finally, if
            the underlying board's _stop() function is called, if present, to do any additional board-specific
            cleanup.
            """
            for motor in motors:
                self.set_motor_speed(motor, 0)
            for servo in servos:
                self.disable_servo(servo)
            if callable(getattr(self, '_stop', None)):
                self._stop(**kwargs)

        @property
        def config(self):
            """
            Read out the servo, motor and ADC configuration as a dict
            """
            result = {}
            if ADCS in self._config:
                result[ADCS] = {index: a.divisor for index, a in self._config[ADCS].items()}
            if MOTORS in self._config:
                result[MOTORS] = {index: {'invert': m.invert} for index, m in self._config[MOTORS].items()}
            if SERVOS in self._config:
                result[SERVOS] = {index: {'pulse_min': s.pulse_min,
                                          'pulse_max': s.pulse_max} for index, s in self._config[SERVOS].items()}
            return result

        @config.setter
        def config(self, d):
            """
            Set the servo, motor, and ADC configuration from a dict
            """
            if ADCS in d and ADCS in self._config:
                for index, divisor in d[ADCS].items():
                    if index in self._config[ADCS]:
                        setattr(self, f'adc{index}_divisor', divisor)
                    else:
                        LOGGER.warning(f'config contained ADC divisor for invalid index {index}')
            if MOTORS in d and MOTORS in self._config:
                for index, invert in d[MOTORS].items():
                    if index in self._config[MOTORS]:
                        setattr(self, f'm{index}_invert', invert)
                    else:
                        LOGGER.warning(f'config contained motor invert for invalid index {index}')
            if SERVOS in d and SERVOS in self._config:
                for index, servo in d[SERVOS].items():
                    if index in self._config[SERVOS]:
                        pulse_min = servo['pulse_min'] if 'pulse_min' in servo else None
                        pulse_max = servo['pulse_max'] if 'pulse_max' in servo else None
                        setattr(self, f's{index}_config', (pulse_min, pulse_max))
                    else:
                        LOGGER.warning(f'config contained servo configuration for invalid index {index}')

        @property
        def motors(self):
            """
            An array of motor indices. For a typical two-motor board this might be [0,1] or similar, in which case
            it would correspond to the injected m0, motor0, m1, motor1 etc properties
            """
            return motors

        @property
        def servos(self):
            """
            An array of servo indices, these correspond to the sXX, servoXX and sXX_config properties
            """
            return servos

        @property
        def adcs(self):
            """
            An array of ADC channel indices, corresponding to the adcXX and adcXX_divisor properties
            :return:
            """
            return adcs

        @property
        def config_yaml(self):
            """
            Config dict as a YAML string
            """
            return yaml.dump(self.config)

    # Set up configuration dict, we only add top level keys if the corresponding facility is requested
    config = {}
    if motors:
        config[MOTORS] = {}
    if servos:
        config[SERVOS] = {}
    if adcs:
        config[ADCS] = {}
    if leds:
        config[LEDS] = {}

    # Inject mXX, motorXX, mXX_invert, and motorXX_invert properties
    for motor in motors:
        m = Motor(motor=motor, invert=False, board=board)
        config['motors'][motor] = m
        for prefix in ['m', 'motor']:
            setattr(Board, f'{prefix}{motor}', property(fget=m.get_value, fset=m.set_value))
            setattr(Board, f'{prefix}{motor}_invert', property(fset=m.set_invert, fget=m.get_invert))

    # Inject sXX, servoXX, sXX_config, and servoXX_config properties
    for servo in servos:
        s = Servo(servo=servo, pulse_min=500, pulse_max=2500, board=board)
        config['servos'][servo] = s
        for prefix in ['s', 'servo']:
            setattr(Board, f'{prefix}{servo}', property(fset=s.set_value, fget=s.get_value))
            setattr(Board, f'{prefix}{servo}_config', property(fset=s.set_config, fget=s.get_config))

    # Inject adcXX and adcXX_divisor properties
    for adc in adcs:
        a = ADC(adc=adc, divisor=default_adc_divisor, board=board)
        config[ADCS][adc] = a
        setattr(Board, f'adc{adc}', property(fget=a.get_value))
        setattr(Board, f'adc{adc}_divisor', property(fset=a.set_divisor, fget=a.get_divisor))

    # Inject ledXX, ledXX_brightness, ledXX_gamma, and ledXX_saturation properties
    for led in leds:
        l = LED(led=led, board=board)
        config[LEDS][led] = l
        setattr(Board, f'led{led}', property(fget=l.get_colour, fset=l.set_colour))
        setattr(Board, f'led{led}_brightness', property(fget=l.get_brightness, fset=l.set_brightness))
        setattr(Board, f'led{led}_gamma', property(fget=l.get_gamma, fset=l.set_gamma))
        setattr(Board, f'led{led}_saturation', property(fget=l.get_saturation, fset=l.set_saturation))

    # Set the supplied object's class to the newly created subclass
    board._config = config
    board.__class__ = Board



# -*- coding: future_fstrings -*-

import logging

from approxeng.hwsupport.util import check_range

LOGGER = logging.getLogger(name='approxeng.hwsupport.motors')
MOTORS = 'motors'


class Motor:
    """
    Holds configuration for a motor. You won't use this class directly.
    """

    def __init__(self, motor, invert, board):
        self.motor = motor
        self.invert = invert
        self.board = board
        self.value = None

    def set_value(self, _, value):
        self.board.set_motor_speed(motor=self.motor, speed=value)

    def get_value(self, _):
        return self.value

    def set_invert(self, _, value):
        if value is None or not isinstance(value, bool):
            raise ValueError(f'm{self.motor}_invert must be True|False, was {value}')
        self.invert = value
        if self.value is not None:
            self.board.set_motor_speed(motor=self.motor, speed=self.value)

    def get_invert(self, _):
        return self.invert


class SetMotorsMixin:
    """
    Mixed into the new class used for the augmented instance to provide the set_motor_speed method
    """

    def set_motor_speed(self, motor: int, speed: float, **kwargs):
        """
        Set a motor speed

        :param motor:
            The motor to set, this must be a value in the array of motor indices
        :param speed:
            Speed from -1.0 to 1.0, values outside this range will be clamped to it
        :param kwargs:
            Any additional arguments to be passed to the underlying _set_motor_speed method defined on the original
            instance pre-augmentation
        :raises:
            ValueError if motors are defined but the supplied index isn't in the array, or no motors are defined.
        """
        if MOTORS in self._config:
            LOGGER.debug(f'set motor m{motor}={speed}')
            if motor not in self._config[MOTORS]:
                raise ValueError(f'motor m{motor} not in {list(self._config[MOTORS].keys())}')
            speed = check_range(speed)
            config = self._config[MOTORS][motor]
            config.value = speed
            self._set_motor_speed(motor, speed if not config.invert else -speed, **kwargs)
        else:
            raise ValueError(f'board has no motor functions, unable to set m{motor}')



# -*- coding: future_fstrings -*-

import logging

LOGGER = logging.getLogger(name='approxeng.hwsupport.util')


def check_range(i):
    """
    Accepts a number, returns that number clamped to a range of -1.0 to 1.0, as a float
    :param i:
        Number
    :return:
        Float between -1.0 and 1.0
    """
    f = float(i)
    if f < -1.0:
        LOGGER.warning('Value < -1.0, returning -1.0')
        return -1.0
    if f > 1.0:
        LOGGER.warning('Value > 1.0, returning 1.0')
        return 1.0
    return f


def check_positive(i):
    f = float(i)
    if f < 0.0:
        LOGGER.warning('Value < 0, returning 0')
        return 0.0
    if f > 1.0:
        LOGGER.warning('Value > 1.0, returning 1.0')
        return 1.0
    return f



import logging

from approxeng.hwsupport.util import check_range

LOGGER = logging.getLogger(name='approxeng.hwsupport.servos')
SERVOS = 'servos'


class Servo:
    """
    Holds configuration for a servo pin
    """

    def __init__(self, servo, pulse_min, pulse_max, board):
        self.servo = servo
        self.pulse_max = pulse_max
        self.pulse_min = pulse_min
        self.value = None
        self.board = board

    def set_value(self, _, value):
        if value is not None:
            self.board.set_servo(servo=self.servo, position=value)
        else:
            self.board.disable_servo(servo=self.servo)

    def get_value(self, _):
        return self.value

    def set_config(self, _, value):
        new_pulse_min, new_pulse_max = value
        if new_pulse_min is not None and not isinstance(new_pulse_min, int):
            raise ValueError(f'pulse_min must be None or int, was {new_pulse_min}')
        if new_pulse_max is not None and not isinstance(new_pulse_max, int):
            raise ValueError(f'pulse_max must be None or int, was {new_pulse_max}')
        self.pulse_min = new_pulse_min or self.pulse_min
        self.pulse_max = new_pulse_max or self.pulse_max
        # PiGPIO won't allow values <500 or >2500 for this, so we clamp them here
        self.pulse_max = min(self.pulse_max, 2500)
        self.pulse_min = max(self.pulse_min, 500)
        if self.value is not None:
            # If we have an active value set then update based on the new
            # configured pulse min / max values
            self.set_value(_, self.value)

    def get_config(self, _):
        return self.pulse_min, self.pulse_max


class SetServosMixin:
    """
    Mixed into the new class used for the augmented instance to provide the set_servo and disable_servo methods
    """

    def _check_servo_index(self, servo):
        """
        Check that a servo is valid, raising ValueError if not and returning the config otherwise.
        """
        if SERVOS not in self._config:
            raise ValueError(f'board has no servo functions, unable to set  / disable s{servo}')
        if servo not in self._config[SERVOS]:
            raise ValueError(f'servo s{servo} not in {list(self._config[SERVOS].keys())}')
        return self._config[SERVOS][servo]

    def set_servo(self, servo: int, position: float, **kwargs):
        """
        Set a servo value

        :param servo:
            The servo to set, this must be a value in the array of servos
        :param position:
            Position from -1.0 (minimum PWM duty cycle) to 1.0 (max PWM). Values outside this range will be clamped to
            it
        :param kwargs:
            Any additional arguments to provide to the underlying _set_servo_pulsewidth method
        :raises:
            ValueError if the supplied servo index isn't available, or there are no servos defined for this board,
            or the position supplied is not a non-None int or float value
        """
        if position is None:
            raise ValueError(f'position for s{servo} must not be None')
        if not isinstance(position, (float, int)):
            raise ValueError(f's{servo} value must be float, was {position}')
        LOGGER.debug(f'set servo s{servo}={position}')
        config = self._check_servo_index(servo)
        position = check_range(position)
        pulse_min, pulse_max = config.pulse_min, config.pulse_max
        config.value = position
        position = -position
        scale = float((pulse_max - pulse_min) / 2)
        centre = float((pulse_max + pulse_min) / 2)
        self._set_servo_pulsewidth(servo, int(centre - scale * position), **kwargs)

    def disable_servo(self, servo: int, **kwargs):
        """
        Set a servo PWM duty cycle to 0, releasing control

        :param servo:
            The servo to set, this must be a value in the array of servos
        :param kwargs:
            Any additional arguments to provide to the underlying _set_servo_pulsewidth method
        :raises:
            ValueError if the supplied servo index isn't available, or there are no servos defined for this board
        """
        LOGGER.debug(f'disable servo s{servo}')
        config = self._check_servo_index(servo)
        config.value = None
        self._set_servo_pulsewidth(servo, 0, **kwargs)



from setuptools import setup, find_namespace_packages

setup(
    name='approxeng.hwsupport',
    version='0.1.7',
    author='Tom Oinn',
    author_email='tomoinn@gmail.com',
    url='https://github.com/ApproxEng/approxeng.hwsupport',
    description='Python robot hardware driver support',
    classifiers=['Programming Language :: Python :: 3.6'],
    packages=find_namespace_packages(),
    install_requires=['future-fstrings', 'pyyaml'],
)
